/*
 * ppplicznik -- counter of ppp connections with many features
 * Copyright (C) 2001 Artur Siekielski <gruesome@go2.pl>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#pragma implementation

#include <netdb.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <iostream>
#include <fstream>
#include <cstdlib>
#include <cctype>
#include <vector>
#include <set>
#include <ctime>
#include <iomanip>
#include <sstream>

#include "koszty.h"
#include "main.h"
#include "ppp.h"
#include "loc.h"

using namespace std;

inline int konfiguracja::coIleTeraz()
{
	return regula_na_teraz->dlugosc_impulsu;
}

void konfiguracja::ustawReguleNaTeraz()
{
	for (vector< pair<int, int> >::const_iterator it = lista_swiat.begin();
			it != lista_swiat.end(); ++it) {
		if (it->first == akt.tm_mday && it->second == akt.tm_mon + 1) {
			regula_na_teraz = &swieto;
			return;
		}
	}
	for (vector<rule>::iterator it = reguly.begin(); it != reguly.end(); ++it) {
		set<int> godziny;
		set<int> dni;
		for(int godz = it->od_godziny; ; ++godz) {
			godziny.insert(godz % 24);
			if ((godz + 1) % 24 == it->do_godziny)
				break;
		}
		for (int dz = it->od_dnia; ; ++dz) {
			dni.insert(dz % 7);
			if (dz % 7 == it->do_dnia)
				break;
		}
		if (dni.find(akt.tm_wday) != dni.end() && godziny.find(akt.tm_hour) != godziny.end()) {
			regula_na_teraz = &(*it);
			return;
		}
	}
	throw error(gettext("No rule for now!"));
}

void konfiguracja::ustawMape()
{
	S["add"]=&konfiguracja::czytajDolicz;
	S["beg_month"]=&konfiguracja::czytajBegMon;
	S["colors"]=&konfiguracja::czytajKolory;
	S["const_hours"]=&konfiguracja::czytajHours;
	S["disconnect"]=&konfiguracja::czytajRozlacz;
	S["disconnect_when"]=&konfiguracja::czytajRozlaczKiedy;
	S["holidays"]=&konfiguracja::czytajSwieta;
	S["file"]=&konfiguracja::czytajPlik;
	S["graph"]=&konfiguracja::czytajWykres;
	S["hostip"]=&konfiguracja::czytajHostIp;
	S["limit"]=&konfiguracja::czytajLimit;
	S["lock_file"]=&konfiguracja::czytajLock;
	S["run"]=&konfiguracja::czytajUruchom;
	S["show"]=&konfiguracja::czytajCoWysw;
	S["tax"]=&konfiguracja::czytajPodatek;
	S["term_options"]=&konfiguracja::czytajTerm;
	dni["monday"] = 1;
	dni["tuesday"] = 2;
	dni["wednesday"] = 3;
	dni["thursday"] = 4;
	dni["friday"] = 5;
	dni["saturday"] = 6;
	dni["sunday"] = 0;
}
void konfiguracja::parsujPlik()
{
	beg_month = 1;
	doliczane = 0;
	gorna_przerwa = 5;
	hostip = false;
	hours = 0;
	ile_podatku = 1;
	ile_sr = 3;
	kolory_dane[0] = 7;
	kolory_dane[1] = 5;
	kolory_dane_bold = false;
	kolory_napisy[0] = 7;
	kolory_napisy[1] = 12;
	kolory_napisy_bold = false;
	kolory_ramka[0] = 7;
	kolory_ramka[1] = 4;
	kolory_ramka_bold = false;
	kolory_wyk_od[0] = 1;
	kolory_wyk_od[1] = 0;
	kolory_wyk_od_bold = true;
	kolory_wyk_wys[0] = 2;
	kolory_wyk_wys[1] = 0;
	kolory_wyk_wys_bold = true;
	limit = false;
	plik_lock = "/tmp/ppplicznik.lock";
	rozlacz_enabled = false;
	rozlacz = false;
	rozlacz_kiedy = 10;
	run = false;
	stt = false;
	wyk_max = 15;
	wyk_one = 650;
	wykres = false;
	//////////////////////////////////////////
	int licznik = 0;
	fstream plik(plik_konf.c_str(), ios::in);
	if (!plik) {
		ostringstream o;
		o << gettext("Error while reading configuration file") << plik_konf << '\n';
		throw error(o.str());
	}
	string line;
	while (getline(plik, line)) {
		++licznik;
		istringstream potok(line);
		string slowo;
		potok >> ws >> slowo;
		if (slowo[0] == '#' || slowo == "" || slowo == "#")
			continue;
		try {
			if (slowo == "rules_begin") {
				while(true) {
					getline(plik, line);
					++licznik;
					if (line.find("rules_end") != string::npos)
						break;
					slowo += line;
				}
				czytajZasady(slowo);
				continue;
			}
			map<string, wskm>::const_iterator it = S.find(slowo);
			if (it == S.end())
				throw error(gettext("Parse error in line "));
			else
				(this->*(it->second))(string(line, slowo.size() + 1));
		}
		catch(error& slowo) {
			plik.close();
			ostringstream duzy;
			duzy << gettext("Error while reading configuration file\n") << slowo.what() << licznik;
			throw error(duzy.str());
		}
	}
	plik.close();
}

void konfiguracja::czytajZasady(const string& wejscie)
{
	istringstream potok(wejscie);
	string slowo;
	potok >> slowo;
	rule temp;
	string dz1;
	string dz2;
	while(true) {
		potok >> dz1 >> dz2 >> temp.od_godziny >> temp.do_godziny >> temp.koszt_impulsu >> temp.dlugosc_impulsu;
		if (potok.eof())
			return;
		if (!potok)
			throw error(gettext("Bad rule definition in line "));
		map<string, int>::const_iterator it = dni.find(dz1);
		if (it == dni.end())
			throw error(gettext("Bad day name in line "));
		temp.od_dnia = it->second;
		it = dni.find(dz2);
		if (it == dni.end())
			throw error(gettext("Bad day name in line "));
		temp.do_dnia = it->second;
		reguly.push_back(temp);
	}
}

void konfiguracja::czytajSwieta(const string& wejscie)
{
	istringstream potok(wejscie);
	potok >> swieto.koszt_impulsu >> swieto.dlugosc_impulsu;
	string data;
	while(true) {
		potok >> data;
		size_t gdzie = data.find('/');
		if (gdzie == string::npos)
			throw error(gettext("Wrong day definition in line "));
		string dzien = data.substr(0, gdzie);
		string miesiac = data.substr(gdzie + 1, string::npos);
		stringstream SS;
		SS << dzien << " " << miesiac;
		int dz, m;
		SS.seekp(ios::beg);
		SS >> dz >> m;
		lista_swiat.push_back(make_pair(dz, m));
		if (potok.eof())
			break;
		if (!potok)
			throw error(gettext("Wrong holidays days definition in line "));
	}
}

void konfiguracja::czytajPodatek(const string& wejscie)
{
	istringstream potok(wejscie);
	int temp;
	potok >> temp;
	if (!potok)
		throw error(gettext("Tax rate badly defined in line "));
	ile_podatku = 1 + (float(temp) / float(100));
}

void konfiguracja::czytajLock(const string& wejscie)
{
	istringstream potok(wejscie);
	potok >> plik_lock;
	if (!potok)
		throw error(gettext("Wrong lock file definition in line "));
}

void konfiguracja::czytajHours(const string& wejscie)
{
	istringstream potok(wejscie);
	potok >> hours >> hour_cost;
	if (!potok || hours < 0 || hour_cost < 0)
		throw error(gettext("Bad hour definition in line "));
}

void konfiguracja::czytajBegMon(const string& wejscie)
{
	istringstream potok(wejscie);
	potok >> beg_month;
	if (!potok || beg_month <=0 || beg_month > 31)
		throw error(gettext("Bad day definition in line "));
}

void konfiguracja::czytajTerm(const string& wejscie)
{
	istringstream potok(wejscie);
	potok >> gorna_przerwa;
	if (!potok || gorna_przerwa < 3)
		throw error(gettext("Bad number of empty lines definition (must be > 2) in line "));
	potok >> wyk_max >> wyk_one >> ile_sr;
	if (!potok)
		throw error(gettext("Bad definition in line "));
	polaczenie::pepepe.ustawTablice();
}

void konfiguracja::czytajWykres(const string&)
{
	wykres = true;
}

void konfiguracja::czytajLimit(const string& wejscie)
{
	limit = true;
	istringstream potok(wejscie);
	potok >> limit_ile;
	if (!potok)
		throw error(gettext("Wrong limit definition in line "));
}

void konfiguracja::czytajCoWysw(const string& wejscie)
{
	cowysw.clear();
	istringstream potok(wejscie);
	int liczba;
	while(true) {
		potok >> liczba;
		if (!potok)
			break;
		if (liczba > 13 || liczba <= 0)
			continue;
		cowysw.push_back(liczba);
	}
}

void konfiguracja::czytajKoloryHelp(istringstream& i, int& kolor, bool& bold)
{
	int liczba;
	i >> liczba;
	if (!i || liczba > 15)
		throw error(gettext("Wrong limit definition in line "));
	if (liczba > 7) {
		liczba -= 8;
		bold = true;
	}
	kolor = liczba;
}

void konfiguracja::czytajKolory(const string& wejscie)
{
	istringstream potok(wejscie);
	czytajKoloryHelp(potok, kolory_ramka[0], kolory_ramka_bold);
	czytajKoloryHelp(potok, kolory_ramka[1], kolory_ramka_bold);
	czytajKoloryHelp(potok, kolory_napisy[0], kolory_napisy_bold);
	czytajKoloryHelp(potok, kolory_napisy[1], kolory_napisy_bold);
	czytajKoloryHelp(potok, kolory_dane[0], kolory_dane_bold);
	czytajKoloryHelp(potok, kolory_dane[1], kolory_dane_bold);
	czytajKoloryHelp(potok, kolory_wyk_od[0], kolory_wyk_od_bold);
	czytajKoloryHelp(potok, kolory_wyk_od[1], kolory_wyk_od_bold);
	czytajKoloryHelp(potok, kolory_wyk_wys[0], kolory_wyk_wys_bold);
	czytajKoloryHelp(potok, kolory_wyk_wys[1], kolory_wyk_wys_bold);
	czytajKoloryHelp(potok, kolory_limit[0], kolory_limit_bold);
	czytajKoloryHelp(potok, kolory_limit[1], kolory_limit_bold);
}

void konfiguracja::czytajHostIp(const string& wejscie)
{
	hostip = true;
}

void konfiguracja::czytajPlik(const string& wejscie)
{
	stt = true;
	istringstream potok(wejscie);
	string word;
	potok >> word;
	if (!potok)
		throw error(gettext("Wrong file definition in line "));
	struct stat ata;
	if (stat(word.c_str(), &ata) < 0) {
		t_mod = time(0);
		return;
	}
	t_mod = ata.st_mtime;
}

void konfiguracja::czytajDolicz(const string& wejscie)
{
	istringstream potok(wejscie);
	potok >> doliczane;
	if (!potok)
		throw error(gettext("Wrong time definition in line "));
}

void konfiguracja::czytajUruchom(const string& wejscie)
{
	run = true;
	istringstream ciag(wejscie);
	char zwiastun = 'a';
	vector<string> wyrazy;
	string slowo;
	while (!isdigit(zwiastun)) {
		slowo.erase();
		ciag >> slowo >> ws;
		if (!ciag)
			throw error(gettext("Wrong time definition in line "));
		wyrazy.push_back(slowo);
		zwiastun = ciag.peek();
	}
	ciag >> ur_przed;
	co_uru = new char *[wyrazy.size() + 1];
	int j = 0;
	for (vector<string>::iterator i = wyrazy.begin(); i != wyrazy.end(); i++, j++) {
		co_uru[j] = const_cast<char*>(i->c_str());
	}
	co_uru[wyrazy.size()] = 0;
	pierwszy_ur = wyrazy[0];
}

void konfiguracja::czytajRozlacz(const string& wejscie)
{
	rozlacz = true;
	istringstream ciag(wejscie);
	vector<string> wyrazy;
	string slowo;
	while (true) {
		slowo.erase();
		ciag >> slowo;
		if (!ciag)
			break;
		if (slowo != "")
			wyrazy.push_back(slowo);
	}
	co_uru_rozl = new char*[wyrazy.size() + 1];
	int j = 0;
	for (vector<string>::iterator it = wyrazy.begin(); it != wyrazy.end(); it++, j++) {
		co_uru_rozl[j] = const_cast<char*>(it->c_str());
	}
	co_uru_rozl[wyrazy.size()] = 0;
	if (wyrazy.size() != 0)
		pierwszy_ur_rozl = wyrazy[0];
}

void konfiguracja::czytajRozlaczKiedy(const string& wejscie)
{
	istringstream potok(wejscie);
	potok >> rozlacz_kiedy;
	if (!potok)
		throw error(gettext("Wrong disconnecting time definition in line "));
}
		
void polaczenie::uruchom()
{
	if (!konf.run || konf.pierwszy_ur == "")
		return;
	int pid = fork();
	if (!pid) {
		if ((execvp(konf.pierwszy_ur.c_str(), konf.co_uru)) < 0 )
			execl("/bin/false", "/bin/false", 0);
	}
}

void polaczenie::uruchom_rozl()
{
	if (!konf.rozlacz || konf.pierwszy_ur_rozl == "")
		return;
	int pid = fork();
	if (!pid) {
		if ((execvp(konf.pierwszy_ur_rozl.c_str(), konf.co_uru_rozl)) < 0)
			execl("/bin/false", "/bin/false", 0);
	}
}

void konfiguracja::czytajZmienne()
{
	char* temp = getenv("IPLOCAL");
	if (temp == NULL || hostip == false) {
		zmienna_ip = gettext("no info");
		zmienna_host = gettext("no info");
		return;
	}
	zmienna_ip = temp;
	hostent* inf;
	in_addr adres;
	if (!inet_aton(temp, &adres)) {
		zmienna_host = gettext("no info");
		return;
	}
	if ((inf = gethostbyaddr(&adres, sizeof adres, AF_INET)) == NULL) {
		zmienna_host = gettext("no info");
		return;
	}
	zmienna_host = inf->h_name;
}

konfiguracja polaczenie::konf;
pppst polaczenie::pepepe;

void polaczenie::oblicz()
{
	trwanie = koniec - poczatek;
	float stare = impulsy;
	impulsy = trwanie / konf.coIleTeraz() + 1;
	koszt = (impulsy * konf.zwrocKoszt()) * konf.ile_podatku;
	if (stare != impulsy) {
		polaczenie::konf.km_dzien += konf.zwrocKoszt() * konf.ile_podatku;
		polaczenie::konf.km_mies += konf.zwrocKoszt() * konf.ile_podatku;
	}
	do_nastepnego = konf.coIleTeraz() * impulsy - trwanie;
	if (konf.run) {
		if (do_nastepnego == konf.ur_przed)
			uruchom();
	}
	if (konf.rozlacz && do_nastepnego == konf.rozlacz_kiedy && konf.rozlacz_enabled)
		uruchom_rozl();
}


#ifndef SYSCONFDIR
#define SYSCONFDIR "/etc"
#endif

konfiguracja::konfiguracja()
{
	plik_konf = SYSCONFDIR"/ppplicznik.conf";
	bez_sprawdzenia = false;
	run = false;
	time_t sj = time(0);
	localtime_r(&sj, &akt);
	wy = miesiac;
	ustawMape();
}

void polaczenie::konwertuj()
{
	localtime_r(&poczatek, &poczatek_tm);
	localtime_r(&koniec, &koniec_tm);
}


polaczenie::polaczenie()
{;
}

ostream& operator<<(ostream& strumien, const polaczenie& connection)
{
	if (connection.konf.wy == konfiguracja::miesiac) {
		if (connection.poczatek_tm.tm_mon != connection.konf.akt.tm_mon
		    || connection.poczatek_tm.tm_year != connection.konf.akt.tm_year)
			return strumien;
	}
	char bufor[128];
	strftime(bufor, 128, "%a %d %b %Y, %H:%M:%S", &connection.poczatek_tm);
	int min = (connection.koniec - connection.poczatek) / 60;
	int sek = (connection.koniec - connection.poczatek) % 60;
#if __GNUC__ < 3
	strumien.form("%20s%6d:%02d%6d%6.2f%8.1f%8.1f\n", bufor, min, sek,
			connection.impulsy, connection.koszt,
			connection.pepepe.bajty_wyslane / 1024, connection.pepepe.bajty_odebrane / 1024);
#else
	strumien << setw(20) << setfill(' ') << bufor << setw(6) << min << setw(0) << ":"
		<< setw(2) << setfill('0') << sek << setw(6) << setfill(' ') << connection.impulsy
		<< setiosflags(ios::fixed) << setiosflags(ios::showpoint) 
		<< setw(8) << setprecision(2) << connection.koszt << setw(8) << setprecision(1)
		<< connection.pepepe.bajty_wyslane / 1024 << setw(8) << setprecision(1)
		<< connection.pepepe.bajty_odebrane / 1024 << '\n';
#endif
	return strumien;
}

fstream& operator<<(fstream& plik, const polaczenie& connection)
{
	plik << connection.poczatek << " " << connection.
		koniec << " " << connection.impulsy << " " << connection.
		koszt << " " << connection.pepepe.bajty_wyslane << " " << connection.pepepe.bajty_odebrane << endl;
	return plik;
}

fstream& operator>>(fstream& plik, polaczenie& connection)
{
	plik >> connection.poczatek >> ws >> connection.koniec >> ws >> connection.
		impulsy >> ws >> connection.koszt >> ws >> connection.pepepe.
		bajty_wyslane >> ws >> connection.pepepe.bajty_odebrane;
	plik.ignore(1);
	connection.trwanie = connection.koniec - connection.poczatek;
	connection.konwertuj();
	return plik;
}
